from typing import Tuple

import yaml

from config_path import ConfigPath

CONFIG = ConfigPath('starling_server', 'rjlyon.com', ".json")

CONFIG_FOLDER = CONFIG.saveFolderPath()
TOKENS_FOLDER = CONFIG.saveFolderPath() / "tokens"
TOKENS_FILENAME = "config.yaml"
TOKENS_FILEPATH = CONFIG_FOLDER / TOKENS_FILENAME


def get_tokens() -> dict:
    try:
        with open(TOKENS_FILEPATH) as file:
            return yaml.safe_load(file)
    except:
        raise Exception(f"Can't find starling config file '{TOKENS_FILEPATH}'")


def get_token_account__from_name(name: str) -> Tuple[str, str]:
    tokens = get_tokens()
    if name.lower() in tokens:
        token = tokens[name.lower()]["token"]
        account_id = tokens[name.lower()]["account_id"]
        return token, account_id

    raise KeyError(f"No token for '{name}'")



# Transaction.py
# Richard Lyon, 20 February 2022
from datetime import datetime
from typing import Optional, List

from pydantic.main import BaseModel


class SourceAmount(BaseModel):
    """Currency dcit"""

    currency: str
    minorUnits: int

    def compute_amount(self, direction: str) -> float:
        """Compute a transaction amount in pounds. Outflow is negative."""
        if direction == "OUT":
            sign = -1
        else:
            sign = 1

        return sign * self.minorUnits / 100.0


class StarlingTransaction(BaseModel):
    """Response from the starling api"""

    feedItemUid: str
    transactionTime: datetime
    counterPartyName: str
    direction: str
    sourceAmount: SourceAmount
    reference: Optional[str] = None
    status: str


class FeedItemsResponse(BaseModel):
    feedItems: List[StarlingTransaction]



from datetime import datetime
from typing import List

from pydantic.main import BaseModel


class StarlingAccount(BaseModel):
    """Represents a Starling Account."""
    accountUid: str
    accountType: str
    defaultCategory: str
    currency: str
    createdAt: datetime
    name: str


class AccountsResponse(BaseModel):
    accounts: List[StarlingAccount]



import re
from typing import TypeVar, Type, Any

import httpx
from pydantic import parse_obj_as

T = TypeVar("T")


async def get(
        token: str, path: str, params: dict = None, return_type: Type[T] = Any
) -> T:
    """Get an api call."""
    API_BASE_URL = "https://api.starlingbank.com/api/v2"

    headers = {"Authorization": f"Bearer {token}", "User-Agent": "python"}
    url = f"{API_BASE_URL}{path}"

    async with httpx.AsyncClient() as client:
        try:
            r = await client.get(url, headers=headers, params=params)
            r.raise_for_status()
        except httpx.HTTPError as e:
            raise Exception(e)

        if return_type is not None:
            return parse_obj_as(return_type, r.json())
        else:
            return r.json()


def clean_string(the_string: str) -> str:
    """Replace multiple spaces with a single space."""
    try:
        return re.sub(" +", " ", the_string)

    except:
        return the_string



from typing import List, Optional

from fastapi import FastAPI

from starling_server.config import get_token_account__from_name
from starling_server.models import Transaction, Account
from starling_server.starling.api import api_get_accounts, api_get_balance_for_account_uid, \
    api_get_settled_transactions_for_account_days

app = FastAPI()


# Accounts #############################################################################


@app.get("/accounts/{name}", tags=["Accounts"], response_model=List[Account])
async def get_accounts(name: str) -> List[Account]:
    token, _ = get_token_account__from_name(name)
    data = await api_get_accounts(token)
    return [Account.from_starling_account(d, token) for d in data.accounts]


@app.get("/account/{name}/balance", tags=["Accounts"], response_model=Optional[float])
async def get_account_balance(name: str) -> Optional[float]:
    token, account_uid = get_token_account__from_name(name)
    return await api_get_balance_for_account_uid(token, account_uid)


# Transactions #############################################################################


@app.get("/account/{name}/transactions", tags=["Transactions"], response_model=List[Transaction])
async def get_settled_transactions_for_account(name: str) -> List[Transaction]:
    return await get_settled_transactions_for_account_days(name, 7)


@app.get(
    "/account/{name}/transactions/{days}",
    tags=["Transactions"],
    response_model=List[Transaction],
)
async def get_settled_transactions_for_account_days(
        name: str, days: int
) -> List[Transaction]:
    token, account_uid = get_token_account__from_name(name)
    return await api_get_settled_transactions_for_account_days(token, account_uid, days)



# cli.py - Command Line Interface
# Richard Lyon, 20 Feb 2022

import os
import shutil
from datetime import datetime
import asyncio


import yaml
from cleo import Command, Application

from starling_server.config import CONFIG, CONFIG_FOLDER, TOKENS_FOLDER, TOKENS_FILEPATH
from starling_server.starling.api import api_get_accounts


class InitCommand(Command):
    """
    Initialise Starling Server configuration file from access tokens

    init
    """
    def handle(self):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.handle_async())

    async def handle_async(self):
        if not self.confirm("WARNING: This will reset configuration data. Continue?", False):
            return

        os.makedirs(CONFIG.saveFolderPath(), exist_ok=True)

        # get tokens from files
        if not TOKENS_FOLDER.is_dir():
            os.mkdir(TOKENS_FOLDER)
            self.line(f"<error>no tokens folder</error>")
            self.line(f"<error>created tokens folder at {TOKENS_FOLDER}</error>")
            self.line("<error>create a file for each token in tokens folder and rerun 'init'</error>")
            self.line("Done")
            return

        tokens = []
        tokens_files = os.listdir(TOKENS_FOLDER)
        for token_file in tokens_files:
            file_path = TOKENS_FOLDER / token_file
            file = open(file_path, "r")
            tokens.append(file.read().strip())
            file.close()

        # build a dictionary of Starling account name and id for each token
        token_dict = {}
        for token in tokens:
            account_info = await api_get_accounts(token)
            for account in account_info.accounts:
                token_dict[account.name.lower()] = {"account_id": account.accountUid, "token": token}

        # save to yaml file, backing up any previous version
        if TOKENS_FILEPATH.is_file():
            backup_name = TOKENS_FILEPATH.stem + "_" + datetime.now().strftime("%Y%m%d%H%M%S") + ".yaml"
            backup_filepath = CONFIG_FOLDER / backup_name
            shutil.copy(TOKENS_FILEPATH, backup_filepath)
            self.line(f"<info>Backed up tokens file to {backup_filepath}</info>")

        with open(TOKENS_FILEPATH, 'w') as file:
            yaml.dump(token_dict, file)
            self.line(f"<info>Created tokens file at {TOKENS_FILEPATH}</info>")

        self.line("Done.")

class ServerCommand(Command):
    """
    Start the Starling Server

    server
    """

    def handle(self):
        from starling_server.main import app
        import uvicorn
        uvicorn.run(app, host="0.0.0.0", port=8080)


cli_app = Application()
cli_app.add(InitCommand())
cli_app.add(ServerCommand())

def cli():
    cli_app.run()


if __name__ == "__main__":
    cli()



from datetime import datetime, timedelta
from typing import Optional, List
from urllib.error import HTTPError

from pydantic import PydanticTypeError

from starling_server.models import Transaction
from starling_server.starling.models.StarlingAccount import AccountsResponse
from starling_server.starling.models.StarlingTransaction import FeedItemsResponse
from starling_server.starling.utils import get


async def api_get_accounts(token: str) -> AccountsResponse:
    return await get(token, "/accounts", None, AccountsResponse)


async def api_get_balance_for_account_uid(token: str, account_uid: str) -> Optional[float]:
    path = f"/accounts/{account_uid}/balance"
    data: dict = await get(token, path, None, dict)

    balance = data.get("clearedBalance")
    if balance is not None:
        minor_units = balance.get("minorUnits")

        if minor_units is not None:
            return minor_units / 100.0

    return None


async def api_get_settled_transactions_for_account_days(token: str, account_uid: str, days: int) -> List[Transaction]:
    max_transaction_timestamp = datetime.now()
    min_transaction_timestamp = max_transaction_timestamp - timedelta(days=days)

    path = f"/feed/account/{account_uid}/settled-transactions-between"
    params = {
        "minTransactionTimestamp": min_transaction_timestamp.strftime(
            "%Y-%m-%dT00:00:00.000Z"
        ),
        "maxTransactionTimestamp": max_transaction_timestamp.strftime(
            "%Y-%m-%dT00:00:00.000Z"
        ),
    }

    try:
        data = await get(token, path, params, FeedItemsResponse)

    except HTTPError:
        raise RuntimeError(f"failed to get transactions for Starling account '{name}'")
    except PydanticTypeError:
        raise RuntimeError(f"")

    return [Transaction.from_starling_transaction(item) for item in data.feedItems]



from datetime import datetime

from pydantic.main import BaseModel

from starling_server.starling.models import StarlingAccount


class Account(BaseModel):
    """Represents an Account."""

    token: str
    uuid: str
    type: str
    default_category: str
    currency: str
    created: datetime
    name: str

    @staticmethod
    def from_starling_account(data: StarlingAccount, token: str) -> "Account":
        return Account(
            token=token,
            uuid=data.accountUid,
            type=data.accountType,
            default_category=data.defaultCategory,
            currency=data.currency,
            created=data.createdAt,
            name=data.name,
        )

    def __str__(self):
        return f"<ACCOUNT: {self.name} {self.uuid}>"



from datetime import datetime
from typing import Optional

from pydantic.main import BaseModel

from starling_server.starling.models import StarlingTransaction
from starling_server.starling.utils import clean_string


class Transaction(BaseModel):
    """Represents a transaction."""

    uuid: str
    time: datetime
    counterparty_name: str
    amount: float
    reference: Optional[str] = None
    status: str

    @staticmethod
    def from_starling_transaction(other: StarlingTransaction) -> "Transaction":
        return Transaction(
            uuid=other.feedItemUid,
            time=other.transactionTime,
            counterparty_name=other.counterPartyName,
            amount=other.sourceAmount.compute_amount(other.direction),
            reference=clean_string(other.reference) if other.reference else None,
            status=other.status,
        )

    def __str__(self):
        return f"<TRANSACTION {self.time} {self.counterparty_name}:{self.amount}>"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['starling_server',
 'starling_server.cli',
 'starling_server.models',
 'starling_server.starling',
 'starling_server.starling.models']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'Sphinx>=4.4.0,<5.0.0',
 'cleo>=0.8.1,<0.9.0',
 'config-path>=1.0.2,<2.0.0',
 'fastapi>=0.74.0,<0.75.0',
 'httpx>=0.22.0,<0.23.0',
 'pydantic>=1.9.0,<2.0.0',
 'sphinx-rtd-theme>=1.0.0,<2.0.0',
 'uvicorn>=0.17.5,<0.18.0']

entry_points = \
{'console_scripts': ['starling-server = starling_server.cli.cli:cli']}

setup_kwargs = {
    'name': 'starling-server',
    'version': '0.1.1',
    'description': 'An API for working with a validated subset of the Starling Bank API',
    'long_description': None,
    'author': 'Richard',
    'author_email': 'richlyon@mac.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
